#include "lz4.h"

// Compression and decompression format based on the HDF5 LZ4 compression filter
// plugin Some parts were copyied and/or adapted from the HDF5 LZ4 compression
// filter plugin source code The copied source code was distributed under a
// BSD-style license, see COPYING next to this file, with the following
// copyright notice:
// HDF5 LZ4 compression filter plugin Copyright 2013-2015 by
// The HDF Group.

// Definitions from LZ4 filter
#define DEFAULT_BLOCK_SIZE 1<<30; /* 1GB. LZ4 needs blocks < 1.9GB. */


int get_lz4_block_info(
        int n_bytes, int* block_size, int* n_blocks, int* header_size)
{
    if (*block_size < 0) {
        return INVALID_BLOCK_SIZE;
    }
    if (*block_size == 0) {
        *block_size = DEFAULT_BLOCK_SIZE;
    }

    if (*block_size > n_bytes)
    {
        *block_size = n_bytes;
    }

    *n_blocks = (n_bytes - 1)/(*block_size) + 1;
    // TODO: implement multi-block messages
    // With the default block size, a single block is sufficient for messages
    // up to 1 GB
    if (*n_blocks != 1) {
        return INVALID_NUMBER_OF_BLOCKS;
    }

    *header_size = 12 + (*n_blocks)*4;

    return 0;
}


struct HDF5ChunkLZ4Serializer: public HDF5ChunkSerializer {
    int get_buffer_size(
        const struct SeedeeNDArray* array, const void* compression_opts, int n_opts)
    {
        if (compression_opts != NULL) {
            return INVALID_COMPRESSION_OPTS;
        }

        int size = array->itemsize * n_elements(array);
        assert(array->size == size);

        // Reserve additional bytes for the HDF5 header
        int block_size = 0;  // use default block size
        int n_blocks;
        int header_size;
        int err = get_lz4_block_info(size, &block_size, &n_blocks, &header_size);
        if (err < 0) {
            return err;
        }

        int total_size = LZ4_COMPRESSBOUND(size) + header_size;

        return total_size;
    }

    int serialize_ndarray(
        const struct SeedeeNDArray* array,
        const void* compression_opts, int n_opts,
        char* buf, int buf_len, char** meta)
    {
        if (compression_opts != NULL) {
            return INVALID_COMPRESSION_OPTS;
        }
        // TODO: Skip check?
        int buffer_size = get_buffer_size(
            array, compression_opts, n_opts
        );
        if (buf_len < buffer_size)
        {
            return BUFFER_TOO_SMALL;
        }

        // Calculate block size
        // TODO: get block size from compression options
        // TODO: Implement multi-block compression
        int block_size = 0;  // use default block size
        int n_blocks;
        int header_size;
        int err = get_lz4_block_info(array->size, &block_size, &n_blocks, &header_size);
        if (err < 0) {
            return err;
        }

        // For head see http://www.hdfgroup.org/services/filters/HDF5_LZ4.pdf
        // First 8 bytes is the uncompressed size
        write_be<uint64_t>(buf, array->size);

        // Next 4 bytes are the uncompressed size per block
        write_be<uint32_t>((char*)buf + 8, block_size);

        // Serialize array
        // TODO: Loop for multi-byte compression
        assert(header_size == 16);
        int compressed_size = LZ4_compress_default(
            array->data, (char*)buf + 16, block_size, buf_len - 16);

        if (compressed_size == 0)
        {
            printf("LZ4 compression failed with error %i\n", compressed_size);
            return COMPRESSION_FAILED;
        }

        if(compressed_size >= block_size)
        // compression did not save any space, do a memcpy instead
        {
            compressed_size = block_size;
            memcpy((char*)buf + 16, array->data, block_size);
        }

        // Fill last 4 bytes of the header, i.e., the compressed block size
        write_be<uint32_t>((char*)buf + 12, compressed_size);

        // Create metadata
        json j = basic_metadata();
        j["data_format"] = FORMAT_HDF5_CHUNK;
        j["compression"] = COMPRESSION_LZ4;
        j["compression_opts"] = {0};  // 0 is the default block size
        json array_json;
        err = array_to_json(array, array_json);
        if (err != 0) {
            return err;
        }
        j["array"] = array_json;

        json_dump_c_str(j, meta);

        return compressed_size + header_size;
    }

    int get_data_size(
        const json& j, const char* buf, int buf_len, bool* zero_copy, struct SeedeeNDArray* array)
    {
        // Compression opts are ignored for decompression but it is probably a
        // good idea to check them for plausibility
        std::vector<int> compression_opts = j.at("compression_opts");
        if (compression_opts.size() != 1 or compression_opts[0] < 0)
        {
            return INVALID_COMPRESSION_OPTS;
        }
        auto j_array = j.at("array");
        int ret = json_to_array(j_array, array);

        // TODO: zero_copy
        *zero_copy = false;

        return ret;
    }


    int deserialize_ndarray(
        const json& j, const char* buf, int buf_len,
        bool zero_copy, struct SeedeeNDArray* array)
    {
        // TODO: zero_copy?
        // TODO: skip check?
        if (array->size != j.at("array").at("size"))
        {
            return BUFFER_TOO_SMALL;
        }

        // Check the original size
        size_t nbytes_uncomp = read_be<uint64_t>(buf);
        assert(array->size == nbytes_uncomp);

        // Read the uncompressed block size from the header.
        size_t block_size_uncomp = read_be<uint32_t>(buf + 8);
        if(block_size_uncomp > nbytes_uncomp)
        {
            block_size_uncomp = nbytes_uncomp;
        }

        // Deserialize array
        // TODO: Loop for multi-byte compression
        // Read the compressed block size
        size_t compressed_block_size = read_be<uint32_t>(buf + 12);

        if (compressed_block_size == block_size_uncomp) {
            // There was no compression
            memcpy(array->data, buf + 16, compressed_block_size);
        }
        else
        {
            int nbytes_consumed = LZ4_decompress_safe(
                buf + 16, array->data, compressed_block_size, array->size);

            if (nbytes_consumed <= 0)
            {
                printf("LZ4 decompression failed with error %i\n", nbytes_consumed);
                return COMPRESSION_FAILED;
            }
        }
        std::vector<int> shape = j.at("array").at("shape");
        assert(shape.size() == array->ndims);
        memcpy(array->shape, shape.data(), array->ndims * sizeof(*(array->shape)));

        return array->size;
    }
};
