
#include "seedee/seedee.h"
#include <nlohmann/json.hpp>
#include <cstdio>
#include <string>
#include <vector>
#include <stdexcept>

#ifndef CMAKE_PROJECT_VERSION
#define CMAKE_PROJECT_VERSION "unknown"
#endif

const char seedee_version[32] = CMAKE_PROJECT_VERSION;

using json = nlohmann::json;

// map SeedeeDataFormat values to JSON as strings
NLOHMANN_JSON_SERIALIZE_ENUM(SeedeeDataFormat, {
    {FORMAT_INVALID, nullptr},
    {FORMAT_HDF5_CHUNK, "hdf5_chunk"},
})

// map SeedeeCompressionType values to JSON as strings
NLOHMANN_JSON_SERIALIZE_ENUM(SeedeeCompressionType, {
    {COMPRESSION_INVALID, nullptr},
    {COMPRESSION_NONE, (int)COMPRESSION_NONE},
    {COMPRESSION_BITSHUFFLE, (int)COMPRESSION_BITSHUFFLE},
    {COMPRESSION_LZ4, (int)COMPRESSION_LZ4},
    {COMPRESSION_DEFLATE, (int)COMPRESSION_DEFLATE},
})


json basic_metadata(void) {
    json j;
    j["protocol"] = "seedee";
    j["version_major"] = 1;
    j["version_minor"] = 0;
    return j;
}


int check_protocol(const json& j) {
    std::string protocol;
    try {
        protocol = j.at("protocol");
    }
    catch (const json::out_of_range& ex)
    {
        // Catch this error here in order to return INVALID_PROTOCOL instead of
        // a general INCOMPLETE_METADATA
        return INVALID_PROTOCOL;
    }
    if (protocol != "seedee")
    {
        return INVALID_PROTOCOL;
    }
    if (j.at("version_major") != 1)
    {
        return INVALID_MAJOR_VERSION;
    }
    if (j.at("version_minor") != 0)
    {
        return INVALID_MINOR_VERSION;
    }
    return 0;
}


struct unknown_data_format: public std::invalid_argument {
    unknown_data_format(std::string msg) : invalid_argument(msg) {};
};

struct unsupported_compression_type: public std::invalid_argument {
    unsupported_compression_type(std::string msg) : invalid_argument(msg) {};
};

struct SeedeeSerializer {
    virtual int get_buffer_size(
        const struct SeedeeNDArray* array, const void* compression_opts, int n_opts) = 0;
    virtual int serialize_ndarray(
        const struct SeedeeNDArray* array, const void* compression_opts, int n_opts,
        char* buf, int buf_len, char** meta) = 0;
    virtual int get_data_size(
        const json& j, const char* buf, int buf_len, bool* zero_copy,
        struct SeedeeNDArray* array) = 0;
    virtual int deserialize_ndarray(
        const json& j, const char* buf, int buf_len,
        bool zero_copy, struct SeedeeNDArray* array) = 0;
};


struct HDF5ChunkSerializer : public SeedeeSerializer {
};



#include "seedee_utils.cpp"
#include "seedee_bitshuffle.cpp"
#include "lz4/seedee_lz4.cpp"
#include "seedee_zlib.cpp"
#include "seedee_no_compression.cpp"


std::unique_ptr<HDF5ChunkSerializer> get_hdf5_chunk_serializer(enum SeedeeCompressionType compression) {
    switch(compression)
    {
        case COMPRESSION_NONE: return std::unique_ptr<HDF5ChunkSerializer>(new HDF5ChunkNoCompressionSerializer());
        case COMPRESSION_BITSHUFFLE: return std::unique_ptr<HDF5ChunkSerializer>(new HDF5ChunkBitshuffleSerializer());
        case COMPRESSION_LZ4: return std::unique_ptr<HDF5ChunkSerializer>(new HDF5ChunkLZ4Serializer());
        case COMPRESSION_DEFLATE: return std::unique_ptr<HDF5ChunkSerializer>(new HDF5ChunkZlibSerializer());
        // TODO: Proper error reporting
        default: throw unsupported_compression_type("Unsupported compression type");
    }
}


std::unique_ptr<SeedeeSerializer> get_serializer(
        enum SeedeeDataFormat data_format, enum SeedeeCompressionType compression) {
    switch(data_format)
    {
        case FORMAT_HDF5_CHUNK: return get_hdf5_chunk_serializer(compression);
        // TODO: Proper error reporting
        default: throw unknown_data_format("Unknown data format");
    }
}


int _seedee_get_buffer_size_impl(
    const struct SeedeeNDArray* array, enum SeedeeDataFormat data_format,
    enum SeedeeCompressionType compression, const void* compression_opts, int n_opts) noexcept
{
    std::unique_ptr<SeedeeSerializer> serializer;
    try {
        serializer = get_serializer(data_format, compression);
    }
    catch(const unknown_data_format& e ) {
        return INVALID_FORMAT;
    }
    catch(const unsupported_compression_type& e ) {
        return INVALID_COMPRESSION;
    }
    return serializer->get_buffer_size(array, compression_opts, n_opts);
}


int seedee_get_buffer_size(
    const struct SeedeeNDArray* array, enum SeedeeDataFormat data_format,
    enum SeedeeCompressionType compression, const void* compression_opts, int n_opts)
{
    return _seedee_get_buffer_size_impl(
        array, data_format, compression, compression_opts, n_opts);
}


int _seedee_serialize_ndarray_impl(
    const struct SeedeeNDArray* array, enum SeedeeDataFormat data_format,
    enum SeedeeCompressionType compression, const void* compression_opts, int n_opts,
    char* buf, int buf_len, char** meta) noexcept
{
    std::unique_ptr<SeedeeSerializer> serializer;
    try {
        serializer = get_serializer(data_format, compression);
    }
    catch(const unknown_data_format& e ) {
        return INVALID_FORMAT;
    }
    catch(const unsupported_compression_type& e ) {
        return INVALID_COMPRESSION;
    }
    return serializer->serialize_ndarray(
            array, compression_opts, n_opts, buf, buf_len, meta);
}


int seedee_serialize_ndarray(
    const struct SeedeeNDArray* array, enum SeedeeDataFormat data_format,
    enum SeedeeCompressionType compression, const void* compression_opts, int n_opts,
    char* buf, int buf_len, char** meta)
{
    return _seedee_serialize_ndarray_impl(
        array, data_format, compression, compression_opts, n_opts, buf, buf_len, meta);
}


int _seedee_get_data_size_impl(
    const char* meta, const char* buf, int buf_len, bool* zero_copy, struct SeedeeNDArray* array) noexcept
{
    json j;
    try {
        j = json::parse(meta);
    }
    catch (json::parse_error& ex)
    {
        return INVALID_JSON;
    }

    int err;
    try {
        err = check_protocol(j);
    }
    catch (const json::out_of_range& ex)
    {
        return INCOMPLETE_METADATA;
    }
    if (err < 0) {
        return err;
    }

    enum SeedeeDataFormat data_format;
    enum SeedeeCompressionType compression;
    try {
        data_format = j.at("data_format").get<SeedeeDataFormat>();
        compression = j.at("compression").get<SeedeeCompressionType>();
    }
    catch (const json::out_of_range& ex)
    {
        return INCOMPLETE_METADATA;
    }

    std::unique_ptr<SeedeeSerializer> serializer;
    try {
        serializer = get_serializer(data_format, compression);
    }
    catch(const unknown_data_format& e ) {
        return INVALID_FORMAT;
    }
    catch(const unsupported_compression_type& e ) {
        return INVALID_COMPRESSION;
    }

    try {
        return serializer->get_data_size(
            j, buf, buf_len, zero_copy, array);
    }
    catch (json::out_of_range& ex)
    {
        return INCOMPLETE_METADATA;
    }
}


int seedee_get_data_size(
    const char* meta, const char* buf, int buf_len, bool* zero_copy, struct SeedeeNDArray* array)
{
    return _seedee_get_data_size_impl(
        meta, buf, buf_len, zero_copy, array
    );
}


int _seedee_deserialize_ndarray_impl(
    const char* meta, const char* buf, int buf_len,
    bool zero_copy, struct SeedeeNDArray* array) noexcept
{
    // TODO: Skip check?
    json j;
    try {
        j = json::parse(meta);
    }
    catch (json::parse_error& ex)
    {
        return INVALID_JSON;
    }

    int err;
    try {
        err = check_protocol(j);
    }
    catch (const json::out_of_range& ex)
    {
        return INCOMPLETE_METADATA;
    }
    if (err < 0) {
        return err;
    }

    enum SeedeeDataFormat data_format;
    enum SeedeeCompressionType compression;
    try {
        data_format = j.at("data_format").get<SeedeeDataFormat>();
        compression = j.at("compression").get<SeedeeCompressionType>();
    }
    catch (json::out_of_range& ex)
    {
        return INCOMPLETE_METADATA;
    }

    std::unique_ptr<SeedeeSerializer> serializer;
    try {
        serializer = get_serializer(data_format, compression);
    }
    catch(const unknown_data_format& e ) {
        return INVALID_FORMAT;
    }
    catch(const unsupported_compression_type& e ) {
        return INVALID_COMPRESSION;
    }
    try {
        return serializer->deserialize_ndarray(
            j, buf, buf_len, zero_copy, array);
    }
    catch (json::out_of_range& ex)
    {
        return INCOMPLETE_METADATA;
    }
}


int seedee_deserialize_ndarray(
    const char* meta, const char* buf, int buf_len,
    bool zero_copy, struct SeedeeNDArray* array)
{
    return _seedee_deserialize_ndarray_impl(
        meta, buf, buf_len, zero_copy, array
    );
}