#include "bitshuffle.h"


struct HDF5ChunkBitshuffleSerializer: public HDF5ChunkSerializer {
    int get_buffer_size(
        const struct SeedeeNDArray* array, const void* compression_opts, int n_opts)
    {
        if (compression_opts != NULL) {
            return INVALID_COMPRESSION_OPTS;
        }
        // Always use the default block size
        // This is equivalent to block_size = 0, but is needed for the HDF5 headers
        size_t block_size = bshuf_default_block_size(array->itemsize);

        // Reserve 12 additional bytes for the HDF5 header
        return bshuf_compress_lz4_bound(
            n_elements(array), array->itemsize, block_size) + 12;
    }

    int serialize_ndarray(
        const struct SeedeeNDArray* array,
        const void* compression_opts, int n_opts,
        char* buf, int buf_len, char** meta)
    {
        if (compression_opts != NULL) {
            return INVALID_COMPRESSION_OPTS;
        }
        // TODO: Skip check?
        int buffer_size = get_buffer_size(
            array, compression_opts, n_opts
        );
        if (buf_len < buffer_size)
        {
            return BUFFER_TOO_SMALL;
        }

        // Always use the default block size
        // This is equivalent to block_size = 0, but is needed for the HDF5 headers
        size_t block_size = bshuf_default_block_size(array->itemsize);

        // For head see http://www.hdfgroup.org/services/filters/HDF5_LZ4.pdf
        // First 8 bytes is the uncompressed size
        write_be<uint64_t>(buf, array->size);

        // Next 4 bytes are the block size
        write_be<uint32_t>((char*)buf + 8, block_size * array->itemsize);

        // Serialize array
        int size = bshuf_compress_lz4(
            array->data, (char*)buf + 12, n_elements(array), array->itemsize, block_size);

        if (size < 0)
        {
            // TODO: how to report the error value to the caller?
            printf("Bitshuffle compression failed with error %i\n", size);
            return COMPRESSION_FAILED;
        }

        // Create metadata
        json j = basic_metadata();
        j["data_format"] = FORMAT_HDF5_CHUNK;
        j["compression"] = COMPRESSION_BITSHUFFLE;
        j["compression_opts"] = {0, 2};
        json array_json;
        int err;
        err = array_to_json(array, array_json);
        if (err != 0) {
            return err;
        }
        j["array"] = array_json;

        json_dump_c_str(j, meta);

        return size + 12;
    }

    int get_data_size(
        const json& j, const char* buf, int buf_len, bool* zero_copy, struct SeedeeNDArray* array)
    {
        // Only default compression opts are supported currently
        std::vector<int> default_compression_opts = {0, 2};
        if (j.at("compression_opts") != default_compression_opts)
        {
            return INVALID_COMPRESSION_OPTS;
        }
        auto j_array = j.at("array");
        int ret = json_to_array(j_array, array);

        // TODO: zero_copy
        *zero_copy = false;

        return ret;
    }


    int deserialize_ndarray(
        const json& j, const char* buf, int buf_len,
        bool zero_copy, struct SeedeeNDArray* array)
    {
        // TODO: zero_copy?
        // TODO: skip check?
        if (array->size != j.at("array").at("size"))
        {
            return BUFFER_TOO_SMALL;
        }

        // Check the original size
        size_t nbytes_uncomp = read_be<uint64_t>(buf);
        assert(array->size == nbytes_uncomp);

        // Read the block size from the header.
        size_t block_size = read_be<uint32_t>(buf + 8) / array->itemsize;

        // Deserialize array
        int nbytes_consumed = bshuf_decompress_lz4(
            buf + 12, array->data, n_elements(j.at("array").at("shape")),
            array->itemsize, block_size);

        if (nbytes_consumed < 0)
        {
            printf("Bitshuffle decompression failed with error %i\n", nbytes_consumed);
            return COMPRESSION_FAILED;
        }

        std::vector<int> shape = j.at("array").at("shape");
        assert(shape.size() == array->ndims);
        memcpy(array->shape, shape.data(), array->ndims * sizeof(*(array->shape)));

        return array->size;
    }
};
