#include "zlib.h"


struct HDF5ChunkZlibSerializer: public HDF5ChunkSerializer {
    int get_buffer_size(
        const struct SeedeeNDArray* array, const void* compression_opts, int n_opts)
    {
        if (compression_opts != NULL) {
            return INVALID_COMPRESSION_OPTS;
        }

        int size = array->itemsize * n_elements(array);
        assert(array->size == size);

        int total_size = compressBound(size);

        return total_size;
    }

    int serialize_ndarray(
        const struct SeedeeNDArray* array,
        const void* compression_opts, int n_opts,
        char* buf, int buf_len, char** meta)
    {
        if (compression_opts != NULL) {
            return INVALID_COMPRESSION_OPTS;
        }
        // TODO: Skip check?
        int buffer_size = get_buffer_size(
            array, compression_opts, n_opts
        );
        if (buf_len < buffer_size)
        {
            return BUFFER_TOO_SMALL;
        }

        // Serialize array
        // Copy buf_len as compress2 will alter it
        uLong dest_len = buf_len;
        int ret = compress2(
            (Bytef*)buf, &dest_len, (Bytef*)array->data, (uLong)array->size, 4);

        if (ret != Z_OK)
        {
            printf("Deflate compression failed with error %i\n", ret);
            return COMPRESSION_FAILED;
        }

        // Create metadata
        json j = basic_metadata();
        j["data_format"] = FORMAT_HDF5_CHUNK;
        j["compression"] = COMPRESSION_DEFLATE;
        // h5py uses an int, but in creation property list it is reported as
        // a list with a single entry
        j["compression_opts"] = {4};
        json array_json;
        int err;
        err = array_to_json(array, array_json);
        if (err != 0) {
            return err;
        }
        j["array"] = array_json;

        json_dump_c_str(j, meta);

        return dest_len;
    }

    int get_data_size(
        const json& j, const char* buf, int buf_len, bool* zero_copy, struct SeedeeNDArray* array)
    {
        // Compression opts are ignored for decompression but it is probably a
        // good idea to check them for plausibility
        std::vector<int> compression_opts = j.at("compression_opts");
        if (compression_opts.size() != 1 or compression_opts[0] < 0 or compression_opts[0] > 9)
        {
            return INVALID_COMPRESSION_OPTS;
        }
        auto j_array = j.at("array");
        int ret = json_to_array(j_array, array);

        // TODO: zero_copy
        *zero_copy = false;

        return ret;
    }

    int deserialize_ndarray(
        const json& j, const char* buf, int buf_len,
        bool zero_copy, struct SeedeeNDArray* array)
    {
        // TODO: zero_copy?
        // TODO: skip check?
        if (array->size != j.at("array").at("size"))
        {
            return BUFFER_TOO_SMALL;
        }

        // Deserialize array
        // Copy buf_len as compress2 will alter it
        uLongf dest_len = array->size;
        int ret = uncompress(
            (Bytef*)array->data, &dest_len, (Byte*) buf, (uLong)buf_len);

        if (ret != Z_OK)
        {
            printf("Delate decompression failed with error %i\n", ret);
            return COMPRESSION_FAILED;
        }
        assert(array->size == dest_len);

        std::vector<int> shape = j.at("array").at("shape");
        assert(shape.size() == array->ndims);
        memcpy(array->shape, shape.data(), array->ndims * sizeof(*(array->shape)));

        return array->size;
    }
};
